import numpy as np
import zengl


def test_blit_array_layer(ctx: zengl.Context):
    array = ctx.image((64, 64), 'rgba8unorm', np.full((3, 64, 64, 4), (200, 100, 0, 255), 'u1'), array=3, levels=2)
    image = ctx.image((64, 64), 'rgba8unorm')

    ctx.new_frame()
    image.clear()
    array.clear_value = (0.0, 0.0, 0.0, 1.0)
    array.face(1).clear()
    array.face(2).blit(image.face())
    ctx.end_frame()
    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [200, 100, 0, 255],
            [200, 100, 0, 255],
            [200, 100, 0, 255],
            [200, 100, 0, 255],
        ],
    )

    ctx.new_frame()
    image.clear()
    array.clear_value = (0.0, 0.0, 1.0, 1.0)
    array.face(2).clear()
    array.face(2).blit(image.face())
    ctx.end_frame()
    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 255, 255],
            [0, 0, 255, 255],
            [0, 0, 255, 255],
            [0, 0, 255, 255],
        ],
    )

    ctx.new_frame()
    array.mipmaps()
    image.clear()
    array.face(layer=2, level=1).blit(image.face(), target_viewport=(0, 0, 32, 32))
    ctx.end_frame()
    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 255, 255],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ],
    )



import pytest
import zengl


def test_invalid_image_size(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.image()

    with pytest.raises(TypeError):
        ctx.image(size=1, format='rgba8unorm')

    with pytest.raises(TypeError):
        ctx.image(size=(1,), format='rgba8unorm')

    with pytest.raises(TypeError):
        ctx.image(size=(1, '2'), format='rgba8unorm')

    with pytest.raises(TypeError):
        ctx.image(size=(2.3, 7.1), format='rgba8unorm')

    with pytest.raises(TypeError):
        ctx.image(size=(1, 2, 3), format='rgba8unorm')


def test_invalid_image_format(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.image(size=(4, 4), format=None)

    with pytest.raises(ValueError):
        ctx.image(size=(4, 4), format='')


def test_invalid_image_data(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.image((1, 1), 'rgba8unorm', data='data')

    with pytest.raises(TypeError):
        ctx.image((1, 1), 'rgba8unorm', data=100)

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', data=b'123')

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', data=b'12345')

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', data=b'1234', samples=2)

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', data=b'12345678', samples=2)

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', data=b'1234', texture=False)


def test_invalid_image_samples(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', samples=0)

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', samples=3)

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', samples=1024)

    with pytest.raises(TypeError):
        ctx.image((1, 1), 'rgba8unorm', samples=2, texture=True)


def test_invalid_image_array(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', array=-1)

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', array=2, data=b'1234')

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', array=2, data=b'1234567')

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', array=2, data=b'123456789')


def test_invalid_image_cubemap(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', cubemap=True, data=b'1234')

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', cubemap=True, data=b'aaaabbbbccccddddeeeefff')

    with pytest.raises(ValueError):
        ctx.image((1, 1), 'rgba8unorm', cubemap=True, data=b'aaaabbbbccccddddeeeefffff')


def test_invalid_image_type(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.image((1, 1), 'rgba8unorm', samples=2, array=2)

    with pytest.raises(TypeError):
        ctx.image((1, 1), 'rgba8unorm', samples=2, cubemap=True)

    with pytest.raises(TypeError):
        ctx.image((1, 1), 'rgba8unorm', array=2, cubemap=True)



import numpy as np
import zengl


def test_render_triangle(ctx: zengl.Context):
    img = ctx.image((256, 256), 'rgba8unorm')
    triangle = ctx.pipeline(
        vertex_shader='''
            #version 330

            vec2 positions[3] = vec2[](
                vec2(0.0, 0.7),
                vec2(-0.85, -0.8),
                vec2(0.85, -0.8)
            );

            void main() {
                gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(1.0, 0.0, 0.0, 1.0);
            }
        ''',
        framebuffer=[img],
        vertex_count=3,
    )
    ctx.new_frame()
    img.clear()
    triangle.render()
    ctx.end_frame()
    pixels = np.frombuffer(img.read(), 'u1').reshape(256, 256, 4)
    x = np.repeat(np.arange(4) * 50 + 50, 4)
    y = np.tile(np.arange(4) * 50 + 50, 4)
    r = [255, 0, 0, 255]
    z = [0, 0, 0, 0]
    np.testing.assert_array_equal(
        pixels[x, y].reshape(4, 4, 4),
        [
            [r, r, r, r],
            [z, r, r, z],
            [z, r, r, z],
            [z, z, z, z],
        ],
    )



import numpy as np
import zengl


def test_image_clear(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'test' * 16)
    assert img.read() == b'test' * 16
    img.clear_value = (0.0, 0.0, 0.0, 1.0)
    img.clear()
    assert img.read() == b'\x00\x00\x00\xff' * 16


def test_image_clear_floats(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rg32float', np.zeros((4, 4, 2), 'f4'))
    content = np.frombuffer(img.read(), 'f4').reshape(4, 4, 2)
    expected = np.zeros((4, 4, 2), 'f4')
    np.testing.assert_array_almost_equal(content, expected, 2)
    img.clear_value = (0.25, 0.75)
    img.clear()
    content = np.frombuffer(img.read(), 'f4').reshape(4, 4, 2)
    expected = np.full((4, 4, 2), (0.25, 0.75), 'f4')
    np.testing.assert_array_almost_equal(content, expected, 2)


def test_image_clear_integers(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32sint', np.zeros((4, 4, 4), 'i4'))
    content = np.frombuffer(img.read(), 'i4').reshape(4, 4, 4)
    expected = np.zeros((4, 4, 4), 'i4')
    np.testing.assert_array_equal(content, expected)
    img.clear_value = (-100, -55, 99, 1048576)
    img.clear()
    content = np.frombuffer(img.read(), 'i4').reshape(4, 4, 4)
    expected = np.full((4, 4, 4), (-100, -55, 99, 1048576), 'i4')
    np.testing.assert_array_equal(content, expected)


def test_invalid_image_clear_cubemap(ctx: zengl.Context):
    img = ctx.image((16, 16), 'rgba8unorm', cubemap=True)
    img.clear_value = (1.0, 0.0, 0.0, 1.0)
    img.clear()
    assert img.face(0).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(1).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(2).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(3).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(4).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(5).read() == b'\xff\x00\x00\xff' * 256
    assert img.read() == b'\xff\x00\x00\xff' * 256 * 6


def test_invalid_image_clear_array(ctx: zengl.Context):
    img = ctx.image((16, 16), 'rgba8unorm', array=4)
    img.clear_value = (1.0, 0.0, 0.0, 1.0)
    img.clear()
    assert img.face(0).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(1).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(2).read() == b'\xff\x00\x00\xff' * 256
    assert img.face(3).read() == b'\xff\x00\x00\xff' * 256
    assert img.read() == b'\xff\x00\x00\xff' * 256 * 4



import pytest
import zengl


def test_missing_vertex_shader(ctx: zengl.Context):
    image = ctx.image((4, 4), 'rgba8unorm')
    with pytest.raises(TypeError):
        ctx.pipeline(
            fragment_shader='''
                #version 330 core
                void main() {
                }
            ''',
            framebuffer=[image],
        )


def test_missing_fragment_shader(ctx: zengl.Context):
    image = ctx.image((4, 4), 'rgba8unorm')
    with pytest.raises(TypeError):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core
                void main() {
                    gl_Position = vec4(1.0);
                }
            ''',
            framebuffer=[image],
        )


def test_missing_framebuffer(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core
                void main() {
                    gl_Position = vec4(1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core
                void main() {
                }
            ''',
        )



import struct

import numpy as np
import zengl


def test(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    uniform_buffer = ctx.buffer(size=64)
    pipeline = ctx.pipeline(
        vertex_shader='''
            #version 330 core

            layout (std140) uniform Common {
                vec4 offset;
                vec4 scale;
                vec4 color;
                int step;
            };

            vec2 positions[3] = vec2[](
                vec2(0.1, 0.0),
                vec2(-0.05, 0.086),
                vec2(-0.05, -0.086)
            );

            void main() {
                gl_Position = vec4(positions[gl_VertexID * step] * scale.xy + offset.xy, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            layout (std140) uniform Common {
                vec4 offset;
                vec4 scale;
                vec4 color;
                int step;
            };

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(color.rgb, 1.0);
            }
        ''',
        layout=[
            {
                'name': 'Common',
                'binding': 2,
            },
        ],
        resources=[
            {
                'type': 'uniform_buffer',
                'binding': 2,
                'buffer': uniform_buffer,
            },
        ],
        framebuffer=[image],
        topology='triangles',
        vertex_count=3,
    )

    uniform_buffer_data = struct.pack(
        '4f4f4f4i',
        *(0.5, 0.5, -999.0, -999.0),
        *(1.0, 1.0, -999.0, -999.0),
        *(0.0, 0.0, 1.0, -999.0),
        *(1, -999, -999, -999),
    )

    ctx.new_frame()
    image.clear()
    uniform_buffer.write(uniform_buffer_data)
    pipeline.render()
    ctx.end_frame()
    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 255, 255],
        ],
    )



import numpy as np
import zengl


def make_pipeline(ctx, framebuffer, color, stencil):
    return ctx.pipeline(
        vertex_shader='''
            #version 330 core

            vec2 positions[3] = vec2[](
                vec2(0.1, 0.0),
                vec2(-0.05, 0.086),
                vec2(-0.05, -0.086)
            );

            void main() {
                gl_Position = vec4(positions[gl_VertexID] + 0.5, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            uniform vec3 color;

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(color, 1.0);
            }
        ''',
        uniforms={
            'color': color,
        },
        stencil={
            'both': stencil,
        },
        depth={
            'func': 'always',
            'write': False,
        },
        framebuffer=framebuffer,
        topology='triangles',
        vertex_count=3,
    )


def test(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    stencil = ctx.image((64, 64), 'depth24plus-stencil8')
    stencil.clear_value = 1.0, 0

    pipeline_1 = make_pipeline(
        ctx,
        [image, stencil],
        (1.0, 0.0, 0.0),
        {
            'fail_op': 'replace',
            'pass_op': 'replace',
            'depth_fail_op': 'replace',
            'compare_op': 'always',
            'compare_mask': 1,
            'write_mask': 1,
            'reference': 1,
        },
    )

    pipeline_2 = make_pipeline(
        ctx,
        [image, stencil],
        (0.0, 1.0, 0.0),
        {
            'fail_op': 'keep',
            'pass_op': 'keep',
            'depth_fail_op': 'keep',
            'compare_op': 'equal',
            'compare_mask': 0xFF,
            'write_mask': 0xFF,
            'reference': 20,
        },
    )

    pipeline_3 = make_pipeline(
        ctx,
        [image, stencil],
        (0.0, 0.0, 1.0),
        {
            'fail_op': 'keep',
            'pass_op': 'keep',
            'depth_fail_op': 'keep',
            'compare_op': 'equal',
            'compare_mask': 0xFF,
            'write_mask': 0xFF,
            'reference': 1,
        },
    )

    ctx.new_frame()
    image.clear()
    stencil.clear()
    pipeline_1.render()
    ctx.end_frame()

    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [255, 0, 0, 255],
        ],
    )

    ctx.new_frame()
    pipeline_2.render()
    ctx.end_frame()

    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [255, 0, 0, 255],
        ],
    )

    ctx.new_frame()
    pipeline_3.render()
    ctx.end_frame()

    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 255, 255],
        ],
    )



from typing import Any, Callable, Dict, Iterable, List, Literal, Protocol, Tuple, TypedDict

CullFace = Literal['front', 'back', 'front_and_back', 'none']
Topology = Literal['points', 'lines', 'line_loop', 'line_strip', 'triangles', 'triangle_strip', 'triangle_fan']

MinFilter = Literal[
    'nearest',
    'linear',
    'nearest_mipmap_nearest',
    'linear_mipmap_nearest',
    'nearest_mipmap_linear',
    'linear_mipmap_linear',
]

MagFilter = Literal['nearest', 'linear']
TextureWrap = Literal['repeat', 'clamp_to_edge', 'mirrored_repeat']
CompareMode = Literal['ref_to_texture', 'none']
CompareFunc = Literal['never', 'less', 'equal', 'lequal', 'greater', 'notequal', 'gequal', 'always']
StencilOp = Literal['zero', 'keep', 'replace', 'incr', 'decr', 'invert', 'incr_wrap', 'decr_wrap']
Step = Literal['vertex', 'instance']

BlendConstant = Literal[
    'zero',
    'one',
    'src_color',
    'one_minus_src_color',
    'src_alpha',
    'one_minus_src_alpha',
    'dst_alpha',
    'one_minus_dst_alpha',
    'dst_color',
    'one_minus_dst_color',
    'src_alpha_saturate',
    'constant_color',
    'one_minus_constant_color',
    'constant_alpha',
    'one_minus_constant_alpha',
    'src1_alpha',
    'src1_color',
    'one_minus_src1_color',
    'one_minus_src1_alpha',
]

VertexFormatShort = Literal[
    '2u1',
    '4u1',
    '2i1',
    '4i1',
    '2nu1',
    '4nu1',
    '2ni1',
    '4ni1',
    '2u2',
    '4u2',
    '2i2',
    '4i2',
    '2nu2',
    '4nu2',
    '2ni2',
    '4ni2',
    '2h',
    '4h',
    '1f',
    '2f',
    '3f',
    '4f',
    '1u',
    '2u',
    '3u',
    '4u',
    '1i',
    '2i',
    '3i',
    '4i',
]

VertexFormat = Literal[
    'uint8x2',
    'uint8x4',
    'sint8x2',
    'sint8x4',
    'unorm8x2',
    'unorm8x4',
    'snorm8x2',
    'snorm8x4',
    'uint16x2',
    'uint16x4',
    'sint16x2',
    'sint16x4',
    'unorm16x2',
    'unorm16x4',
    'snorm16x2',
    'snorm16x4',
    'float16x2',
    'float16x4',
    'float32',
    'float32x2',
    'float32x3',
    'float32x4',
    'uint32',
    'uint32x2',
    'uint32x3',
    'uint32x4',
    'sint32',
    'sint32x2',
    'sint32x3',
    'sint32x4',
]

ImageFormat = Literal[
    'r8unorm',
    'rg8unorm',
    'rgba8unorm',
    'r8snorm',
    'rg8snorm',
    'rgba8snorm',
    'r8uint',
    'rg8uint',
    'rgba8uint',
    'r16uint',
    'rg16uint',
    'rgba16uint',
    'r32uint',
    'rg32uint',
    'rgba32uint',
    'r8sint',
    'rg8sint',
    'rgba8sint',
    'r16sint',
    'rg16sint',
    'rgba16sint',
    'r32sint',
    'rg32sint',
    'rgba32sint',
    'r16float',
    'rg16float',
    'rgba16float',
    'r32float',
    'rg32float',
    'rgba32float',
    'rgba8unorm-srgb',
    'depth16unorm',
    'depth24plus',
    'depth24plus-stencil8',
    'depth32float',
]

Vec3 = Tuple[float, float, float]
Viewport = Tuple[int, int, int, int]
Bytes = bytes | Any

class LayoutBinding(TypedDict, total=False):
    name: str
    binding: int

class BufferResource(TypedDict, total=False):
    type: Literal['uniform_buffer']
    binding: int
    buffer: 'Buffer'
    offset: int
    size: int

class SamplerResource(TypedDict, total=False):
    type: Literal['sampler']
    binding: int
    image: 'Image'
    min_filter: MinFilter
    mag_filter: MagFilter
    min_lod: float
    max_lod: float
    lod_bias: float
    wrap_x: TextureWrap
    wrap_y: TextureWrap
    wrap_z: TextureWrap
    compare_mode: CompareMode
    compare_func: CompareFunc
    max_anisotropy: float

class VertexBufferBinding(TypedDict, total=False):
    buffer: 'Buffer'
    format: VertexFormat
    location: int
    offset: int
    stride: int
    step: Step

class DepthSettings(TypedDict, total=False):
    write: bool
    func: CompareFunc

class StencilFaceSettings(TypedDict, total=False):
    fail_op: StencilOp
    pass_op: StencilOp
    depth_fail_op: StencilOp
    compare_op: CompareFunc
    compare_mask: int
    write_mask: int
    reference: int

class StencilSettings(TypedDict, total=False):
    front: StencilFaceSettings
    back: StencilFaceSettings
    both: StencilFaceSettings

class BlendSettings(TypedDict, total=False):
    enable: bool | int
    src_color: BlendConstant
    dst_color: BlendConstant
    src_alpha: BlendConstant
    dst_alpha: BlendConstant

class Info(TypedDict):
    vendor: str
    renderer: str
    version: str
    glsl: str

class Limits(TypedDict):
    max_uniform_buffer_bindings: int
    max_uniform_block_size: int
    max_combined_uniform_blocks: int
    max_combined_texture_image_units: int
    max_vertex_attribs: int
    max_draw_buffers: int
    max_samples: int

class ImageFace:
    image: 'Image'
    size: Tuple[int, int]
    samples: int
    color: bool
    def clear(self) -> None: ...
    def blit(
        self,
        target: 'ImageFace',
        target_viewport: Viewport | None = None,
        source_viewport: Viewport | None = None,
        filter: bool = True,
        srgb: bool | None = None,
    ) -> None: ...

class ContextLoader(Protocol):
    def load_opengl_function(name: str) -> int: ...

class Buffer:
    size: int
    def write(self, data: Bytes, offset: int = 0) -> None: ...
    def map(self, size: int | None = None, offset: int | None = None, discard: bool = False) -> memoryview: ...
    def unmap(self) -> None: ...

class Image:
    size: Tuple[int, int]
    samples: int
    color: bool
    clear_value: Iterable[int | float] | int | float
    def face(self, layer: int = 0, level: int = 0) -> ImageFace: ...
    def clear(self) -> None: ...
    def write(
        self,
        data: Bytes,
        size: Tuple[int, int] | None = None,
        offset: Tuple[int, int] | None = None,
        layer: int | None = None,
        level: int = 0,
    ) -> None: ...
    def mipmaps(self) -> None: ...
    def read(self, size: Tuple[int, int] | None = None, *, offset: Tuple[int, int] | None = None) -> bytes: ...
    def blit(
        self,
        target: 'Image' | None = None,
        target_viewport: Viewport | None = None,
        source_viewport: Viewport | None = None,
        filter: bool = True,
        srgb: bool | None = None,
    ) -> None: ...

class Pipeline:
    vertex_count: int
    instance_count: int
    first_vertex: int
    viewport: Viewport
    uniforms: Dict[str, memoryview] | None
    def render(self) -> None: ...

class Context:
    info: Info
    limits: Limits
    includes: Dict[str, str]
    before_frame: Callable | None
    after_frame: Callable | None
    frame_time: int
    screen: int
    def buffer(
        self,
        data: Bytes | None = None,
        size: int | None = None,
        dynamic: bool = False,
        index: bool = False,
        uniform: bool = False,
        external: int = 0,
    ) -> Buffer: ...
    def image(
        self,
        size: Tuple[int, int],
        format: ImageFormat,
        data: Bytes | None = None,
        samples: int = 1,
        array: int = 0,
        levels: int = 1,
        texture: bool | None = None,
        cubemap: bool = False,
        external: int = 0,
    ) -> Image: ...
    def pipeline(
        self,
        vertex_shader: str = ...,
        fragment_shader: str = ...,
        layout: Iterable[LayoutBinding] = (),
        resources: Iterable[BufferResource | SamplerResource] = (),
        uniforms: Dict[str, Any] | None = None,
        depth: DepthSettings | None = None,
        stencil: StencilSettings | None = None,
        blend: BlendSettings | None = None,
        framebuffer: Iterable[Image] | None = ...,
        vertex_buffers: Iterable[VertexBufferBinding] = (),
        index_buffer: Buffer | None = None,
        short_index: bool = False,
        cull_face: CullFace = 'none',
        topology: Topology = 'triangles',
        vertex_count: int = 0,
        instance_count: int = 0,
        first_vertex: int = 0,
        viewport: Viewport | None = None,
        uniform_data: memoryview | None = None,
        viewport_data: memoryview | None = None,
        render_data: memoryview | None = None,
        includes: Dict[str, str] | None = None,
    ) -> Pipeline: ...
    def new_frame(self, reset: bool = True, clear: bool = True, frame_time: bool = False) -> None: ...
    def end_frame(self, clean: bool = True, flush: bool = True, sync: bool = False) -> None: ...
    def release(self, obj: Buffer | Image | Pipeline | Literal['shader_cache'] | Literal['all']) -> None: ...

def init(loader: ContextLoader | None = None): ...
def context() -> Context: ...
def inspect(self, obj: Buffer | Image | Pipeline): ...
def camera(
    eye: Vec3,
    target: Vec3,
    up: Vec3 = (0.0, 0.0, 1.0),
    fov: float = 45.0,
    aspect: float = 1.0,
    near: float = 0.1,
    far: float = 1000.0,
    size: float = 1.0,
    clip: bool = False,
) -> bytes: ...
def bind(buffer: Buffer | None, layout: str, *attributes: Iterable[int]) -> List[VertexBufferBinding]: ...
def calcsize(layout: str) -> int: ...
def loader(headless: bool = False) -> ContextLoader: ...
